# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

class _Reasons:
    BACKEND_MISSING_INTERFACE: _Reasons
    UNSUPPORTED_HASH: _Reasons
    UNSUPPORTED_CIPHER: _Reasons
    UNSUPPORTED_PADDING: _Reasons
    UNSUPPORTED_MGF: _Reasons
    UNSUPPORTED_PUBLIC_KEY_ALGORITHM: _Reasons
    UNSUPPORTED_ELLIPTIC_CURVE: _Reasons
    UNSUPPORTED_SERIALIZATION: _Reasons
    UNSUPPORTED_X509: _Reasons
    UNSUPPORTED_EXCHANGE_ALGORITHM: _Reasons
    UNSUPPORTED_DIFFIE_HELLMAN: _Reasons
    UNSUPPORTED_MAC: _Reasons
